/*
  Device driver for 3Com 3C589B/C/D EtherLink III

  T. Nakagawa
  2009-01-11	created for uIP
  2010-04-12	modified for usock
*/


#include <stdint.h>
#include <avr/io.h>
#include <util/delay.h>
#include "device.h"


#define PORT_EL3_COMMAND 0x0e
#define PORT_EL3_STATUS 0x0e
#define PORT_EL3_W0_ADRSCONF 0x06
#define PORT_EL3_W0_RESCONF 0x08
#define PORT_EL3_W0_EEPROMCMD 0x0a
#define PORT_EL3_W0_EEPROMDATA 0x0c
#define PORT_EL3_W1_TXPIODATA 0x00
#define PORT_EL3_W1_RXPIODATA 0x00
#define PORT_EL3_W1_RXSTATUS 0x08
#define PORT_EL3_W1_TIMERTEG 0x0a
#define PORT_EL3_W1_TXSTATUS 0x0b
#define PORT_EL3_W1_FREETRANSBYTES 0x0c
#define PORT_EL3_W2_ADRS0 0x00
#define PORT_EL3_W2_ADRS1 0x02
#define PORT_EL3_W2_ADRS2 0x04
#define PORT_EL3_W4_MEDTYPE 0x0a
#define IO_WR 4
#define IO_RD 5
#define IO_WE 6
#define IO_CE 7


static void outp_pcmcia(unsigned char adrs, unsigned char data) {
  /* address */
  PORTB = _BV(IO_CE) | _BV(IO_WE) | _BV(IO_RD) | _BV(IO_WR) | adrs;
  /* CE#->L */
  PORTB &= ~_BV(IO_CE);
  /* Data */
  DDRD = 0xff;
  PORTD = data;
  /* WE#->L */
  PORTB &= ~_BV(IO_WE);
  _delay_ms(0.35);
  /* WE#->H */
  PORTB |= _BV(IO_WE);
  /* Data */
  PORTD = 0x00;
  DDRD = 0x00;
  /* CE#->H */
  PORTB |= _BV(IO_CE);

  return;
}


static void outpb_el3(unsigned char adrs, unsigned char data) {
  /* address */
  PORTB = _BV(IO_CE) | _BV(IO_WE) | _BV(IO_RD) | _BV(IO_WR) | adrs;
  /* CE1#->L */
  PORTB &= ~_BV(IO_CE);
  /* Data */
  DDRD = 0xff;
  PORTD = data;
  /* IOWR#->L */
  PORTB &= ~_BV(IO_WR);
  _delay_ms(0.2);
  /* IOWR#->H */
  PORTB |= _BV(IO_WR);
  /* Data */
  PORTD = 0x00;
  DDRD = 0x00;
  /* CE1#->H */
  PORTB |= _BV(IO_CE);

  return;
}


static void outpw_el3(unsigned char adrs, unsigned char low, unsigned char high) {
  /* Low byte */

  outpb_el3(adrs + 0, low);
  outpb_el3(adrs + 1, high);

  return;
}


static unsigned char inpb_el3(unsigned char adrs) {
  uint8_t data;

  /* Low byte */

  /* address */
  PORTB = _BV(IO_CE) | _BV(IO_WE) | _BV(IO_RD) | _BV(IO_WR) | adrs;
  /* CE1#->L */
  PORTB &= ~_BV(IO_CE);
  /* IORD#->L */
  PORTB &= ~_BV(IO_RD);
  _delay_ms(0.2);
  /* Data */
  data = PIND;
  /* IORD#->H */
  PORTB |= _BV(IO_RD);
  /* CE1#->H */
  PORTB |= _BV(IO_CE);

  return data;
}


void dev_init(void) {
  uint8_t e0, e1, e2, e3, e4, e5;

  _delay_ms(10);

  /* Initialize PORT */
  PORTB = 0xf0;
  DDRB = 0xff;
  PORTD = 0x00;
  DDRD = 0x00;
  PORTC = 0x30;
  DDRC = 0x0f;

  /* Soft RESET Card */
  outp_pcmcia(0x0, 0x80);
  _delay_ms(1);
  outp_pcmcia(0x0, 0x00);

  /* Enable 8-bit data path */
  outp_pcmcia(0x2, 0x20);

  /* Enable I/O interface */
  outp_pcmcia(0x0, 0x01);

  /* Set ROM SIZE, ROM BASE, and I/O BASE to zero */
  outpw_el3(PORT_EL3_W0_ADRSCONF, 0x00, 0x00);

  /* Disable IRQ */
  outpw_el3(PORT_EL3_W0_RESCONF, 0x00, 0x0f);

  /* Reset transmitter */
  outpw_el3(PORT_EL3_COMMAND, 0x00, 0x0b << 3);
  for (; inpb_el3(PORT_EL3_STATUS + 1) & 0x10; ) ;	/* Wait */

  /* Reset receiver */
  outpw_el3(PORT_EL3_COMMAND, 0x00, 0x05 << 3);

  /* Set read zero mask */
  outpw_el3(PORT_EL3_COMMAND, 0xff, 0x0f << 3);

  /* Set Interrupt Mask (for timer) */
  outpw_el3(PORT_EL3_COMMAND, 0x40, 0x0e << 3);

  /* Set RxFilter */
  outpw_el3(PORT_EL3_COMMAND, 0x05, 0x10 << 3);

  /* Set the station address */
  outpw_el3(PORT_EL3_W0_EEPROMCMD, 0x80 + 0, 0x00);
  _delay_ms(0.2);	/* 162us */
  e0 = inpb_el3(PORT_EL3_W0_EEPROMDATA + 1);
  e1 = inpb_el3(PORT_EL3_W0_EEPROMDATA);
  outpw_el3(PORT_EL3_W0_EEPROMCMD, 0x80 + 1, 0x00);
  _delay_ms(0.2);	/* 162us */
  e2 = inpb_el3(PORT_EL3_W0_EEPROMDATA + 1);
  e3 = inpb_el3(PORT_EL3_W0_EEPROMDATA);
  outpw_el3(PORT_EL3_W0_EEPROMCMD, 0x80 + 2, 0x00);
  _delay_ms(0.2);	/* 162us */
  e4 = inpb_el3(PORT_EL3_W0_EEPROMDATA + 1);
  e5 = inpb_el3(PORT_EL3_W0_EEPROMDATA);

  outpw_el3(PORT_EL3_COMMAND, 0x02, 0x01 << 3);	/* Window 2 */
  outpw_el3(PORT_EL3_W2_ADRS0, e0, e1);
  outpw_el3(PORT_EL3_W2_ADRS1, e2, e3);
  outpw_el3(PORT_EL3_W2_ADRS2, e4, e5);

  /* Enable Link beat and Jabber */
  outpw_el3(PORT_EL3_COMMAND, 0x04, 0x01 << 3);	/* Window 4 */
  outpw_el3(PORT_EL3_W4_MEDTYPE, 0xc0, 0x00);

  /* Stop coaxial tranceiver */
  outpw_el3(PORT_EL3_COMMAND, 0x00, 0x17 << 3);

  /* Enable transmitter */
  outpw_el3(PORT_EL3_COMMAND, 0x00, 0x09 << 3);

  /* Enable receiver */
  outpw_el3(PORT_EL3_COMMAND, 0x00, 0x04 << 3);

  outpw_el3(PORT_EL3_COMMAND, 0x01, 0x01 << 3);	/* Window 1 */

  return;
}


uint16_t dev_recv(uint8_t *data, uint16_t size) {
  uint16_t i;
  uint16_t len;
  uint16_t timer = 0;
  uint8_t t;

  /* Polling for 0.1sec using Timer Register */
  outpw_el3(PORT_EL3_COMMAND, 0x41, 0x0d << 3);	/* Reset the timer */
  outpw_el3(PORT_EL3_COMMAND, 0x40, 0x0c << 3);
  while (!(inpb_el3(PORT_EL3_STATUS) & 0x10)) {	/* Check RX Complete */
    t = inpb_el3(PORT_EL3_W1_TIMERTEG);
    if (t >= 128) {
      timer += t + 2;
      /* Count 0.1 sec using the 3.2us timer */
      if (timer >= (uint16_t)(100000.0 / 3.2)) {
        timer = 0;
        return 0;
      }
      outpw_el3(PORT_EL3_COMMAND, 0x41, 0x0d << 3);	/* Reset the timer */
      outpw_el3(PORT_EL3_COMMAND, 0x40, 0x0c << 3);
    }
  }

  /* Read RX bytes */
  len = inpb_el3(PORT_EL3_W1_RXSTATUS);
  len |= (inpb_el3(PORT_EL3_W1_RXSTATUS + 1) & 0x07) << 8;
  if (len > size) len = size;

  /* Read FIFO */
  for (i = 0; i < size; i++) {
    data[i] = inpb_el3(PORT_EL3_W1_RXPIODATA);
  }

  /* Discard top packet of RX FIFO */
  outpw_el3(PORT_EL3_COMMAND, 0x00, 0x08 << 3);
  for (; inpb_el3(PORT_EL3_STATUS + 1) & 0x10; ) ;	/* Wait */

  return size;
}


void dev_send(uint8_t *data, uint16_t size) {
  uint16_t i;
  uint8_t stat;
  uint16_t len;

  do {
    /* Read TX free bytes */
    len = inpb_el3(PORT_EL3_W1_FREETRANSBYTES);
    len |= inpb_el3(PORT_EL3_W1_FREETRANSBYTES + 1) << 8;
  } while (len <= size + 3);

  while ((stat = inpb_el3(PORT_EL3_W1_TXSTATUS)) & 0x80) {
    if (stat & 30) {	/* TX Reset required */
      outpw_el3(PORT_EL3_COMMAND, 0x00, 0x0b << 3);
      for (; inpb_el3(PORT_EL3_STATUS + 1) & 0x10; ) ;	/* Wait */
      outpw_el3(PORT_EL3_COMMAND, 0x00, 0x09 << 3);
    }
    outpb_el3(PORT_EL3_W1_TXSTATUS, 0x00);	/* Pop the stack */
  }

  /* Write preamble */
  outpb_el3(PORT_EL3_W1_TXPIODATA, (unsigned char)(size & 0x00ff));
  outpb_el3(PORT_EL3_W1_TXPIODATA, (unsigned char)(size >> 8));
  outpb_el3(PORT_EL3_W1_TXPIODATA, 0x00);
  outpb_el3(PORT_EL3_W1_TXPIODATA, 0x00);

  /* Write data (with padding for dword alignment) */
  for (i = 0; i < ((size + 3) & ~0x3); i++) {
    outpb_el3(PORT_EL3_W1_TXPIODATA, *(data++));
  }

  return;
}
