/*
  usock - A Tiny TCP/UDP Socket Library

  T. Nakagawa
  2010-04-06
  2010-04-14	version 0.5
*/


#include <stdint.h>
#include <string.h>
#include "device.h"
#include "config.h"
#include "usock.h"


#define STAT_ACTIVE 0x01
#define STAT_RDOPEN 0x02
#define STAT_WROPEN 0x04

#define MAC_IP 0x0800
#define MAC_ARP 0x0806

#define ARP_REQUEST 1
#define ARP_REPLY 2

#define IP_ICMP 1
#define IP_TCP 6
#define IP_UDP 17

#define ICMP_ECHO_REPLY 0
#define ICMP_ECHO 8     

#define TCP_FIN 0x01
#define TCP_SYN 0x02
#define TCP_RST 0x04
#define TCP_PSH 0x08
#define TCP_ACK 0x10
#define TCP_URG 0x20

#define BUF_SIZE ((42 - 14) + 14 + MTU)
#define MAC_SIZE (6 + 6 + 4 + MTU)
#define MSS (MTU - 40)

#define IPADDREQ(a0, a1) (memcmp(a0, a1, 4) == 0)
#define IPADDRCPY(a0, a1) memcpy(a0, a1, 4);
#define NETWORKEQ(a0, a1, m) ((((a0)[0] & (m)[0]) == ((a1)[0] & (m)[0])) && (((a0)[1] & (m)[1]) == ((a1)[1] & (m)[1])) && (((a0)[2] & (m)[2]) == ((a1)[2] & (m)[2])) && (((a0)[3] & (m)[3]) == ((a1)[3] & (m)[3])))
#define MACADDRCPY(a0, a1) memcpy(a0, a1, 6);


/*
  ARP table
*/
struct Atbl {
  uint8_t ipaddr[4];
  uint8_t macaddr[6];
  uint16_t lifetime;
};

/*
  MAC header
*/
struct Mac {
  uint8_t daddr[6];
  uint8_t saddr[6];
  uint16_t type;
};

/*
  MAC+ARP header
*/
struct Arp {
  struct Mac mac;
  uint16_t htype;
  uint16_t ptype;
  uint16_t hplen;
  uint16_t oper;
  uint8_t smacaddr[6];
  uint8_t sipaddr[4];
  uint8_t dmacaddr[6];
  uint8_t dipaddr[4];
};

/*
  IP header
*/
struct Ip {
  uint8_t verihl;
  uint8_t tos;
  uint16_t len;
  uint16_t id;
  uint16_t flagoffset;
  uint8_t ttl;
  uint8_t proto;
  uint16_t chksum;
  uint8_t saddr[4];
  uint8_t daddr[4];
  uint8_t data[1];
};

/*
  UDP header
*/
struct Udp {
  uint16_t sport;
  uint16_t dport;
  uint16_t len;
  uint16_t chksum;
  uint8_t data[1];
};

/*
  TCP header
*/
struct Tcp {
  uint16_t sport;
  uint16_t dport;
  uint32_t seqno;
  uint32_t ackno;
  uint8_t offset;
  uint8_t flag;
  uint16_t win;
  uint16_t chksum;
  uint16_t urgptr;
  uint8_t data[1];
};


struct Usock_tcb Usock_tcb[TCB_SIZE];
uint8_t Usock_ipaddress[4] = {IPADDR0, IPADDR1, IPADDR2, IPADDR3};
uint8_t Usock_netmask[4] = {NMASK0, NMASK1, NMASK2, NMASK3};
uint8_t Usock_defaultrouter[4] = {DRADDR0, DRADDR1, DRADDR2, DRADDR3};
uint16_t Usock_timeout = TIMEOUT;
uint8_t *Usock_data;
uint16_t Usock_size;
uint16_t Usock_max;

static struct Atbl Atbl[ARPTBL_SIZE];
static uint8_t Buffer[BUF_SIZE];
static struct Arp * const Arp = (struct Arp *)Buffer;
static struct Mac * const Mac = (struct Mac *)(Buffer + (42 - 14));
static struct Ip * const Ip = (struct Ip *)(Buffer + (42 - 14) + 14);
static struct Udp * const Udp = (struct Udp *)(Buffer + (42 - 14) + 14 + 20);
static struct Tcp * const Tcp = (struct Tcp *)(Buffer + (42 - 14) + 14 + 20);
static uint16_t Nextport = 1024;
static uint32_t Isn = 0;
static uint16_t Ipid = 0;
static uint16_t Timer = 0;
static uint16_t Timelimit;
static uint16_t Arptimer = 10;
static uint8_t Macaddress[6] = {MACADDR0, MACADDR1, MACADDR2, MACADDR3, MACADDR4, MACADDR5};


static int8_t checkport(uint8_t mode, uint16_t port);
static uint16_t mac_recv(void);
static int8_t ip_recv(int8_t sd);
static int8_t udp_recv(int8_t sd);
static int8_t tcp_recv(int8_t sd);
static int8_t mac_send(uint16_t len);
static int8_t ip_send(int8_t sd, uint8_t proto, uint16_t len);
static int8_t udp_send(int8_t sd);
static int8_t tcp_send(int8_t sd, uint8_t flag, uint8_t optlen);
static void tcp_getmss(int8_t sd);
static int8_t icmp_echoreply(void);
static uint16_t chksum(uint8_t *mem, uint16_t len, uint16_t proton);
static int8_t arp_request(uint8_t *addr);
static int8_t arp_process(struct Arp *arp);
static void arp_timer(void);


/*
  usock_open
*/
int8_t usock_open(uint8_t mode, uint16_t lport, uint16_t rport, uint8_t *raddr) {
  static uint8_t initflag = 0;
  int8_t sd;

  Timelimit = Timer + Usock_timeout;
  Isn += 65536;

  /* Initialize */
  if (!initflag) {
    dev_init();
    initflag = 1;
  }

  /* Find an unused descripter */
  for (sd = 0; sd < TCB_SIZE; sd++) {
    if (Usock_tcb[sd].stat == 0) break;
  }
  if (sd == TCB_SIZE) return -1;

  /* Check the availability of the port */
  if (lport != 0) {
    if (checkport(mode, lport)) return -1;
  } else {
    while (checkport(mode, Nextport)) if (++Nextport >= 4096) Nextport = 1024;
    lport = Nextport;
  }

  Usock_tcb[sd].mode = mode;
  Usock_tcb[sd].lport = lport;
  Usock_tcb[sd].rport = rport;
  Usock_tcb[sd].sndnxt = Isn;

  if (rport != 0) {	/* Client (active open) */
    IPADDRCPY(Usock_tcb[sd].raddr, raddr);

    /* UDP */

    if (mode == USOCK_UDP) {
      Usock_size = 0;
      Usock_tcb[sd].stat = STAT_ACTIVE;
      return sd;
    }

    /* TCP */

    /* Send SYN & Receive ACK */
    Usock_size = 0;
    Tcp->data[0] = 0x02;
    Tcp->data[1] = 0x04;
    Tcp->data[2] = (Usock_tcb[sd].mss >> 8);
    Tcp->data[3] = (Usock_tcb[sd].mss & 0xff);
    if (tcp_send(sd, TCP_SYN, 4)) return -1;
    do {
      if (tcp_recv(sd)) return -1;
    } while (!(Tcp->flag & TCP_ACK));

    /* Receive SYN */
    while (!(Tcp->flag & TCP_SYN)) {
      if (tcp_recv(sd)) return -1;
    }
    tcp_getmss(sd);
    Usock_tcb[sd].rcvnxt = NTOHL(Tcp->seqno) + 1;

    /* Send ACK */
    Usock_size = 0;
    if (tcp_send(sd, TCP_ACK, 0)) return -1;
  } else {	/* Server (passive open) */
    /* UDP */

    if (mode == USOCK_UDP) {
      if (usock_recv(sd)) return -1;
      Usock_tcb[sd].stat = STAT_ACTIVE;
      return sd;
    }

    /* TCP */

    /* Receive SYN */
    do {
      if (tcp_recv(sd)) return -1;
    } while (!(Tcp->flag & TCP_SYN));
    tcp_getmss(sd);
    Usock_tcb[sd].rcvnxt = NTOHL(Tcp->seqno) + 1;

    /* Send SYN/ACK & Receive ACK */
    Usock_size = 0;
    Tcp->data[0] = 0x02;
    Tcp->data[1] = 0x04;
    Tcp->data[2] = (Usock_tcb[sd].mss >> 8);
    Tcp->data[3] = (Usock_tcb[sd].mss & 0xff);
    if (tcp_send(sd, (TCP_SYN | TCP_ACK), 4)) return -1;

    /* Receive ACK */
    do {
      if (tcp_recv(sd)) return -1;
    } while (!(Tcp->flag & TCP_ACK));
  }

  Usock_tcb[sd].stat = (STAT_ACTIVE | STAT_RDOPEN | STAT_WROPEN);

  return sd;
}


/*
  Close connection
*/
void usock_close(int8_t sd) {
  Timelimit = Timer + Usock_timeout;

  /* UDP */

  if (Usock_tcb[sd].mode == USOCK_UDP) {
    Usock_tcb[sd].stat = 0;

    return;
  }

  /* TCP */

  /* Close the local-to-remote connection */
  if ((Usock_tcb[sd].stat & STAT_WROPEN)) {
    Usock_size = 0;
    usock_send(sd);
  }

  /* Wait for the remote-to-local connection closed */
  if (!(Usock_tcb[sd].stat & STAT_WROPEN)) {
    while ((Usock_tcb[sd].stat & STAT_RDOPEN)) {
      if (tcp_recv(sd)) break;
    }
  }

  Usock_tcb[sd].stat = 0;

  return;
}


/*
  Allocate buffer
*/
void usock_alloc(int8_t sd) {
  if (Usock_tcb[sd].mode == USOCK_UDP) {
    Usock_data = Udp->data;
    Usock_max = MTU - 20 - 8;
  } else {
    Usock_data = Tcp->data;
    Usock_max = MTU - 20 - 20;
  }

  return;
}


/*
  Receive data
*/
int8_t usock_recv(int8_t sd) {
  uint16_t size;

  Timelimit = Timer + Usock_timeout;

  /* UDP */

  if (Usock_tcb[sd].mode == USOCK_UDP) {
    if (udp_recv(sd)) return 1;

    return 0;
  }

  /* TCP */

  do {
    if (!(Usock_tcb[sd].stat & STAT_RDOPEN)) {
      Usock_size = 0;
      return 0;
    }
    if (tcp_recv(sd)) return 1;
  } while (Usock_size == 0);

  /* Send ACK */
  size = Usock_size;
  Usock_size = 0;
  if (tcp_send(sd, TCP_ACK, 0)) return 1;
  Usock_size = size;

  return 0;
}


/*
  Send data
*/
uint16_t usock_send(int8_t sd) {
  uint16_t size;

  Timelimit = Timer + Usock_timeout;

  /* UDP */

  if (Usock_tcb[sd].mode == USOCK_UDP) {
    udp_send(sd);
    Usock_size = 0;

    return 0;
  }

  /* TCP */

  /* Local-to-remote connection is closed if Usock_size == 0 */
  if (Usock_size == 0) {
    /* Send FIN & Receive ACK */
    if (tcp_send(sd, (TCP_FIN | TCP_ACK), 0)) return 0;
    do {
      if (tcp_recv(sd)) return 0;
    } while (!(Tcp->flag & TCP_ACK));

    Usock_tcb[sd].stat &= ~STAT_WROPEN;

    return 0;
  }

  if (!(Usock_tcb[sd].stat & STAT_WROPEN)) return -1;
  if (Usock_size > Usock_tcb[sd].mss) Usock_size = Usock_tcb[sd].mss;
  size = Usock_size;

  /* Send data */
  if (tcp_send(sd, (TCP_PSH | TCP_ACK), 0)) {
    Usock_size = 0;
    return 0;
  }

  /* Receive ACK */
  do {
    if (tcp_recv(sd)) {
      Usock_size = 0;
      return 0;
    }
  } while (!(Tcp->flag & TCP_ACK));

  return size;
}


/*
  Check if a specified port is used 
*/
static int8_t checkport(uint8_t mode, uint16_t port) {
  uint8_t i;

  for (i = 0; i < TCB_SIZE; i++) {
    if (Usock_tcb[i].stat != 0 && Usock_tcb[i].mode == mode && Usock_tcb[i].lport == port) return 1;
  }

  return 0;
}


/*
  Receive MAC data
*/
static uint16_t mac_recv(void) {
  uint16_t len;

  Isn++;

  len = dev_recv((uint8_t *)Mac, MAC_SIZE);
  if (len == 0) return 0;

  if (Mac->type == HTONS(MAC_ARP)) arp_process((struct Arp *)Mac);

  return len;
}


/*
  Receive IP data

  Return value: 0=no error, 1=timeout
*/
static int8_t ip_recv(int8_t sd) {
  uint16_t len;

  for (; ; ) {
    len = mac_recv();
    if (len == 0) {
      Timer++;
      if (Timer == Timelimit) return 1;
      if (Timer == Arptimer) arp_timer();
      continue;
    }

    if (Mac->type != HTONS(MAC_IP)) continue;
    if (Ip->verihl != 0x45) continue;
    if (NTOHS(Ip->len) > len) continue;
    if ((Ip->flagoffset & HTONS(0x3fff)) != 0) continue;
    if (!IPADDREQ(Ip->daddr, Usock_ipaddress)) continue;
    if (chksum((uint8_t *)Ip, 20, 0) != 0xffff) continue;

    if (Ip->proto == IP_ICMP && Ip->data[0] == ICMP_ECHO) {
      if (icmp_echoreply()) return 1;
      continue;
    }

    break;
  }

  return 0;
}


/*
  Receive UDP data

  Return value: 0=no error, 1=timeout
*/
static int8_t udp_recv(int8_t sd) {
  uint16_t len;

  while (!ip_recv(sd)) {
    if (Ip->proto != IP_UDP) continue;
    if (NTOHS(Udp->dport) != Usock_tcb[sd].lport) continue;

    if (Usock_tcb[sd].rport != 0) {
      if (NTOHS(Udp->sport) != Usock_tcb[sd].rport || !IPADDREQ(Ip->saddr, Usock_tcb[sd].raddr)) continue;
    } else {
      Usock_tcb[sd].rport = NTOHS(Udp->sport);
      IPADDRCPY(Usock_tcb[sd].raddr, Ip->saddr);
    }

    len = NTOHS(Udp->len);
    if (Udp->chksum != 0x0000 && chksum((uint8_t *)Udp, len, HTONS(IP_UDP)) != 0xffff) continue;

    Usock_data = Udp->data;
    Usock_size = len - 8;

    return 0;
  }

  return 1;
}


/*
  Receive TCP data

  Return value: 0=no error, 1=timeout
*/
static int8_t tcp_recv(int8_t sd) {
  uint16_t len, inc;

  while (!ip_recv(sd)) {
    if (Ip->proto != IP_TCP) continue;
    if (NTOHS(Tcp->dport) != Usock_tcb[sd].lport) continue;

    if (Usock_tcb[sd].rport != 0) {
      if (NTOHS(Tcp->sport) != Usock_tcb[sd].rport || !IPADDREQ(Ip->saddr, Usock_tcb[sd].raddr)) continue;
    } else {
      Usock_tcb[sd].rport = NTOHS(Tcp->sport);
      IPADDRCPY(Usock_tcb[sd].raddr, Ip->saddr);
      if ((Tcp->flag & TCP_ACK) != 0) {
	Usock_tcb[sd].sndnxt = NTOHL(Tcp->ackno);
	Usock_size = 0;
	tcp_send(sd, TCP_RST, 0);
	Usock_tcb[sd].rport = 0;
	continue;
      }
      if (!(Tcp->flag & TCP_SYN) || (Tcp->flag & (TCP_RST | TCP_FIN))) continue;
    }

    /* Checksum */
    len = NTOHS(Ip->len) - 20;
    if (chksum((uint8_t *)Tcp, len, HTONS(IP_TCP)) != 0xffff) continue;

    /* Usock_connection reset */
    if ((Tcp->flag & TCP_RST)) {
      Usock_tcb[sd].stat &= ~(STAT_RDOPEN | STAT_WROPEN);
      return 1;
    }

    /* Check ACK number */
    if ((Tcp->flag & TCP_ACK) && NTOHL(Tcp->ackno) != Usock_tcb[sd].sndnxt) {
      Usock_size = 0;
      tcp_send(sd, TCP_RST, 0);
      Usock_tcb[sd].stat &= ~(STAT_RDOPEN | STAT_WROPEN);
      return 1;
    }

    /* Check sequence number */
    if (!(Tcp->flag & TCP_SYN) && NTOHL(Tcp->seqno) != Usock_tcb[sd].rcvnxt) {
      /* Request re-transmission */
      Usock_size = 0;
      if (tcp_send(sd, TCP_ACK, 0)) return 1;
      continue;
    }
    len -= ((Tcp->offset >> 4) << 2);
    inc = len;
    if ((Tcp->flag & (TCP_SYN | TCP_FIN))) inc++;
    Usock_tcb[sd].rcvnxt += inc;

    /* Check FIN flag */
    if ((Tcp->flag & TCP_FIN)) {
      Usock_tcb[sd].stat &= ~STAT_RDOPEN;
      Usock_size = 0;
      if (tcp_send(sd, TCP_ACK, 0)) return 1;
    }

    Usock_data = Tcp->data;
    Usock_size = len;

    return 0;
  }

  return 1;
}


/*
  Send MAC data

  Return value: 0=no error, 1=timeout
*/
static int8_t mac_send(uint16_t len) {
  int8_t i;
  uint8_t *addr;

  Isn++;

  if (NETWORKEQ(Ip->daddr, Usock_ipaddress, Usock_netmask)) {
    addr = Ip->daddr;
  } else {
    addr = Usock_defaultrouter;
  }

  for (i = 0; i < ARPTBL_SIZE; i++) {
    if (Atbl[i].lifetime != 0 && IPADDREQ(Atbl[i].ipaddr, addr)) break;
  }
  if (i == ARPTBL_SIZE) {
    i = arp_request(addr);
    if (i < 0) return 1;
  }

  MACADDRCPY(Mac->daddr, Atbl[i].macaddr);
  MACADDRCPY(Mac->saddr, Macaddress);
  Mac->type = HTONS(MAC_IP);

  dev_send((uint8_t *)Mac, 14 + len);

  return 0;
}


/*
  Send IP data
*/
static int8_t ip_send(int8_t sd, uint8_t proto, uint16_t len) {
  Ipid++;

  Ip->proto = proto;
  IPADDRCPY(Ip->saddr, Usock_ipaddress);
  IPADDRCPY(Ip->daddr, Usock_tcb[sd].raddr);

  /* Checksum with pseudo-header */
  if (proto == IP_UDP) {
    Udp->chksum = ~chksum((uint8_t *)Udp, len, HTONS(IP_UDP));
    if (Udp->chksum == 0x0000) Udp->chksum = 0xffff;
  } else if (proto == IP_TCP) {
    Tcp->chksum = ~chksum((uint8_t *)Tcp, len, HTONS(IP_TCP));
  }

  len += 20;
  Ip->verihl = 0x45;
  Ip->tos = 0;
  Ip->len = HTONS(len);
  Ip->id = HTONS(Ipid);
  Ip->flagoffset = 0;
  Ip->ttl = TTL;
  Ip->chksum = 0x0000;

  /* Checksum */
  Ip->chksum = ~chksum((uint8_t *)Ip, 20, 0);

  return mac_send(len);
}


/*
  Send UDP data
*/
static int8_t udp_send(int8_t sd) {
  uint16_t len;

  len = 8 + Usock_size;

  Udp->sport = HTONS(Usock_tcb[sd].lport);
  Udp->dport = HTONS(Usock_tcb[sd].rport);
  Udp->len = HTONS(len);
  Udp->chksum = 0x0000;	/* Caluculated later in ip_send() */

  return ip_send(sd, IP_UDP, len);
}


/*
  Send TCP data
*/
static int8_t tcp_send(int8_t sd, uint8_t flag, uint8_t optlen) {
  uint16_t len, inc;

  len = 20 + optlen + Usock_size;

  Tcp->sport = HTONS(Usock_tcb[sd].lport);
  Tcp->dport = HTONS(Usock_tcb[sd].rport);
  Tcp->seqno = HTONL(Usock_tcb[sd].sndnxt);
  Tcp->ackno = HTONL(Usock_tcb[sd].rcvnxt);
  Tcp->offset = (((20 + optlen) >> 2) << 4);
  Tcp->flag = flag;
  Tcp->win = HTONS(WIN_SIZE);
  Tcp->chksum = 0x0000;	/* Caluculated later int ip_send() */

  if (ip_send(sd, IP_TCP, len)) return 1;

  inc = Usock_size;
  if ((Tcp->flag & (TCP_SYN | TCP_FIN))) inc++;
  Usock_tcb[sd].sndnxt += inc;

  return 0;
}


/*
  Process MSS option of TCP packet
*/
static void tcp_getmss(int8_t sd) {
  uint8_t i;
  uint16_t mss;

  mss = 536;

  /* Process MSS option */
  for (i = 0; i < ((Tcp->offset >> 4) << 2); ) {
    if (Tcp->data[i] == 0x00) {
      break;
    } else if (Tcp->data[i] == 0x01) {
      i++;
    } else if (Tcp->data[i] == 0x02 && Tcp->data[i + 1] == 0x04) {
      mss = (((uint16_t)Tcp->data[i + 2] << 8) | Tcp->data[i + 3]);
      break;
    } else {
      if (Tcp->data[i + 1] == 0) break;
      i += Tcp->data[i + 1];
    }
  }

  if (mss > MSS) mss = MSS;
  Usock_tcb[sd].mss = mss;

  return;
}


/*
  Reply ICMP Echo
*/
static int8_t icmp_echoreply(void) {
  uint8_t tmp;

  Ip->data[0] = ICMP_ECHO_REPLY;
  tmp = Ip->saddr[0]; Ip->saddr[0] = Ip->daddr[0]; Ip->daddr[0] = tmp;
  tmp = Ip->saddr[1]; Ip->saddr[1] = Ip->daddr[1]; Ip->daddr[1] = tmp;
  tmp = Ip->saddr[2]; Ip->saddr[2] = Ip->daddr[2]; Ip->daddr[2] = tmp;
  tmp = Ip->saddr[3]; Ip->saddr[3] = Ip->daddr[3]; Ip->daddr[3] = tmp;

  /* Checksum */
  Ip->data[2] += ICMP_ECHO - ICMP_ECHO_REPLY;
  if (Ip->data[2] < ICMP_ECHO - ICMP_ECHO_REPLY) Ip->data[3]++;

  return mac_send(NTOHS(Ip->len));
}


/*
  Calculate checksum (return value is in network order)
*/
static uint16_t chksum(uint8_t *mem, uint16_t len, uint16_t proton) {
  uint16_t i;
  uint32_t sum;

  if ((len & 0x0001)) {
    mem[len] = 0x00;	/* This operation is safe in this program */
  }
  sum = 0;
  for (i = 0; i < len; i += 2) {
    sum += *(uint16_t *)&mem[i];
  }

  if (proton != 0) {	/* pseudo-header */
    sum += *(uint16_t *)&Ip->saddr[0];
    sum += *(uint16_t *)&Ip->saddr[2];
    sum += *(uint16_t *)&Ip->daddr[0];
    sum += *(uint16_t *)&Ip->daddr[2];
    sum += proton;
    sum += HTONS(len);
  }

  sum = (sum & 0x0000ffff) + (sum >> 16);
  sum = (sum & 0x0000ffff) + (sum >> 16);

  return (uint16_t)sum;
}


/*
  Request ARP
*/
static int8_t arp_request(uint8_t *addr) {
  int8_t i;
  uint16_t len;

  /* Send ARP */
  memset(Arp->mac.daddr, 0xff, 6);
  MACADDRCPY(Arp->mac.saddr, Macaddress);
  Arp->mac.type = HTONS(MAC_ARP);
  Arp->htype = HTONS(0x0001);
  Arp->ptype = HTONS(0x0800);
  Arp->hplen = HTONS(0x0604);
  Arp->oper = HTONS(ARP_REQUEST);
  MACADDRCPY(Arp->smacaddr, Macaddress);
  IPADDRCPY(Arp->sipaddr, Usock_ipaddress);
  memset(Arp->dmacaddr, 0x00, 6);
  IPADDRCPY(Arp->dipaddr, addr);
  dev_send((uint8_t *)Arp, sizeof(struct Arp));

  /* Receive ARP */
  for (; ; ) {
    len = dev_recv((uint8_t *)Arp, sizeof(struct Arp));
    if (len == 0) {
      Timer++;
      if (Timer == Timelimit) return -1;
      if (Timer == Arptimer) arp_timer();
      continue;
    }
    if (Arp->mac.type == HTONS(MAC_ARP)) {
      i = arp_process(Arp);
      if (i >= 0 && IPADDREQ(Atbl[i].ipaddr, addr)) return i;
    }
  }

  return -1;
}


/*
  Process ARP packet
*/
static int8_t arp_process(struct Arp *arp) {
  int8_t i, j;

  for (i = 0, j = 0; i < ARPTBL_SIZE; i++) {
    if (Atbl[i].lifetime != 0 && IPADDREQ(Atbl[i].ipaddr, arp->sipaddr)) break;
    if (Atbl[i].lifetime < Atbl[j].lifetime) j = i;
  }

  if (IPADDREQ(arp->dipaddr, Usock_ipaddress)) {
    /* Register to ARP table */
    if (i == ARPTBL_SIZE) i = j;
    IPADDRCPY(Atbl[i].ipaddr, arp->sipaddr);
    MACADDRCPY(Atbl[i].macaddr, arp->smacaddr);
    Atbl[i].lifetime = ARP_LIFETIME;

    /* Reply for ARP request */
    if (arp->oper == HTONS(ARP_REQUEST)) {
      MACADDRCPY(arp->mac.daddr, arp->smacaddr);
      MACADDRCPY(arp->mac.saddr, Macaddress);
      arp->oper = HTONS(ARP_REPLY);
      MACADDRCPY(arp->dmacaddr, arp->smacaddr);
      IPADDRCPY(arp->dipaddr, arp->sipaddr);
      MACADDRCPY(arp->smacaddr, Macaddress);
      IPADDRCPY(arp->sipaddr, Usock_ipaddress);
      dev_send((uint8_t *)arp, sizeof(struct Arp));
    }
  } else {
    /* Update ARP table */
    if (i != ARPTBL_SIZE) {
      IPADDRCPY(Atbl[i].ipaddr, arp->dipaddr);
      MACADDRCPY(Atbl[i].macaddr, arp->dmacaddr);
      Atbl[i].lifetime = ARP_LIFETIME;
    }
  }

  return (i == ARPTBL_SIZE) ? -1 : i;
}


/*
  Periodic ARP table update
*/
static void arp_timer(void) {
  int8_t i;

  Arptimer = Timer + 1000;

  for (i = 0; i < ARPTBL_SIZE; i++) {
    if (Atbl[i].lifetime != 0) Atbl[i].lifetime--;
  }

  return;
}
